/* $Id$        */
/*--------------------------------------------------------------------*/
/*;  Copyright (C) 2010,2012                                          */
/*;  Associated Universities, Inc. Washington DC, USA.                */
/*;                                                                   */
/*;  This program is free software; you can redistribute it and/or    */
/*;  modify it under the terms of the GNU General Public License as   */
/*;  published by the Free Software Foundation; either version 2 of   */
/*;  the License, or (at your option) any later version.              */
/*;                                                                   */
/*;  This program is distributed in the hope that it will be useful,  */
/*;  but WITHOUT ANY WARRANTY; without even the implied warranty of   */
/*;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    */
/*;  GNU General Public License for more details.                     */
/*;                                                                   */
/*;  You should have received a copy of the GNU General Public        */
/*;  License along with this program; if not, write to the Free       */
/*;  Software Foundation, Inc., 675 Massachusetts Ave, Cambridge,     */
/*;  MA 02139, USA.                                                   */
/*;                                                                   */
/*;Correspondence about this software should be addressed as follows: */
/*;         Internet email: bcotton@nrao.edu.                         */
/*;         Postal address: William Cotton                            */
/*;                         National Radio Astronomy Observatory      */
/*;                         520 Edgemont Road                         */
/*;                         Charlottesville, VA 22903-2475 USA        */
/*--------------------------------------------------------------------*/
#include "ObitVLAGain.h"
#include "ObitUVDesc.h"
#include "ObitUtil.h"

/*----------------Obit: Merx mollis mortibus nuper ------------------*/
/**
 * \file ObitVLAGain.c
 * ObitVLAGain utility function definitions.
 */

/**
 * Private
 * Get polynomial gain curves at a set of frequencies from those at another
 * Polynomials are in terms of zenith angle (deg)
 * \param inTerm     Number of polynomial terms
 * \param inNFreq    Number of input frequencies
 * \param inFreqs    Array of input frequencies (Hz) asc. order
 * \param inPoly     Polynomial curves at inFreqs [freq][term]
 * \param outNFreq   Number of output frequencies
 * \param outFreqs   Array of output frequencies (Hz) asc. order
 * \param outPoly   [out] polynomials at outFreqs
 */
static void GetGainCurves (olong nTerm, 
			   olong inNFreq, odouble *inFreqs, ofloat **inPoly, 
			   olong outNFreq, odouble *outFreqs, ofloat **outPoly)
{
  olong nZA=30;   /* Number of zenith angles tabulated over 0-90 deg */
  olong i, j, iZA, iHi, iLo;
  ofloat ZA, delta, wtLo, wtHi;
  ofloat *ZAs=NULL, *wt=NULL, **curves=NULL, *curveInp=NULL;

  /* If only one, just copy */
  if (inNFreq==1) {
    for (i=0; i<outNFreq; i++) {
      for (j=0; j<nTerm; j++) {
	outPoly[i][j] = inPoly[0][j];
      }
    }
    return;
  } /* end of only one */

  /* work arrays */
  curveInp = g_malloc0(nZA*sizeof(ofloat));
  ZAs      = g_malloc0(nZA*sizeof(ofloat));
  wt       = g_malloc0(nZA*sizeof(ofloat));
  curves   = g_malloc0(inNFreq*sizeof(ofloat*));
  for (i=0; i<inNFreq; i++) curves[i] = g_malloc0(nZA*sizeof(ofloat));

  /* Create input curves */
  delta = 90.0 / nZA;
  for (iZA=0; iZA<nZA; iZA++) {
    ZA       = iZA * delta;     /* Zenith angle */
    ZAs[iZA] = ZA;
    wt[iZA]  = 1.0;             /* Weight */
    for (i=0; i<inNFreq; i++) {
      curves[i][iZA] = inPoly[i][0] + inPoly[i][1]*ZA + inPoly[i][2]*ZA*ZA + inPoly[i][3]*ZA*ZA*ZA;
    }
  }

  /* Interpolate/Fit output curves */
  for (i=0; i<outNFreq; i++) {
    /* Find frequency in input arrays */
    iHi = iLo = 0;  /* In case lower freq than all input */
    for (j=1; j<inNFreq; j++) {
      if ((outFreqs[i]>=inFreqs[j-1]) && (outFreqs[i]<=inFreqs[j])) {
	iHi = j; iLo = j-1; break;
			      }
    } /* end loop find frequency in input */
    /* Check if freq higher than all input */
    if (outFreqs[i]>inFreqs[inNFreq-1]) iHi = iLo = inNFreq-1;

    /* Get interpolation weights */
    if (iHi == iLo) {wtLo = 1.0; wtHi = 0.0;}  /* Same */
    else {  /* not the same */
      wtLo = fabs(inFreqs[iHi] - outFreqs[i]) / fabs(inFreqs[iHi] - inFreqs[iLo]);
      wtHi = 1.0 - wtLo;
    }
    
    /* Interpolate */
    for (iZA=0; iZA<nZA; iZA++) curveInp[iZA] = wtLo*curves[iLo][iZA] + wtHi*curves[iHi][iZA];

    /* Fit */
    FitPoly (outPoly[i], 3, ZAs, curveInp, wt, nZA);

  } /* End loop interpolating/fitting output curves */

  /* Cleanup */
  if (curveInp) g_free(curveInp);
  if (ZAs)      g_free(ZAs);
  if (wt)       g_free(wt);
  if (curves) {
    for (i=0; i<inNFreq; i++) if (curves[i]) g_free(curves[i]);
    g_free(curves);
  }

} /* end GetGainCurves */

/**
 *  Get VLA Gain Curve parameters from list of measurements
 *  Returned as polynomial in zenith angle (deg)
 * \param Ant      Antenna number
 * \param JD       Julian data desired
 * \param Freq     Average frequency (Hz), to get correct band
 * \param nfreq   [number of frequencies to get gains for
 * \param gainFreq Array of frequencies (Hz) to get gains for
 * \param gain     [out] Array of arrays gain polynomial coefficients
 *                 [freq][term] 4 terms implemented
 */
void ObitVLAGainParseGain (olong Ant, odouble JD, odouble Freq, 
			   olong nfreq, odouble *gainFreq, ofloat **gain)
{
  olong i, iarray=-1, n, iline, date, sdate, edate, antno, yr, mo, day;
  ofloat avgGain[5] = {1.0, 0.0,0.0,0.0,0.0};  /* Flat gain if all else fails */
  ofloat **garray=NULL;   /* Gains for up to eight frequencies */
  odouble farray[8] = {0.,0.,0.,0.,0.,0.,0.,0.};     /* Frequencies of gains */
  olong narray=0;        /* Number of frequencies found */
  gboolean found, gotBand;
  /* gain curve file for the EVLA.  bjb.  2012sep20.                     */
  /* the format of the lines looks like:                                 */
  /* bb aa yyy1m1d1 yyy2m2d2 c0 c1 c2 c3                                 */
  /* where:                                                              */
  /* b = band designation (currently one of: 4, P, L, C, X, U, K, A, Q)  */
  /* aa = antenna number.  if aa=0, this is an average over all antennas.*/
  /* yyyy1 m1 d1 = year, month, day before which this gain curve entry is*/
  /*               not valid                                             */
  /* yyyy2 m2 d2 = year, month, day after which this gain curve entry is */
  /*               not valid                                             */
  /* c0, c1, c2, c3 = gain curve coefficients                            */
  /*                                                                     */
  /* the subroutine that reads the lines in handles free format (i.e.,   */
  /* any number of spaces > 0 between the entries).  The date strings    */
  /* are 8 consecutive characters with no blanks                         */
  /*---------------------------------------------------------------------*/
  /**/
  /* 4-band.  all set to 1.0 at all elevations*/
  /**/
  gchar *gainLines[] = {
   "4    75.0  0 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  1 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  2 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  3 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  4 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  5 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  6 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  7 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  8 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0  9 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 10 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 11 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 12 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 13 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 14 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 15 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 16 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 17 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 18 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 19 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 20 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 21 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 22 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 23 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 24 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 25 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 26 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 27 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 28 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "4    75.0 29 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  /**/
  /*-----------------------------------------------------------------------*/
  /**/
  /* P-band.  all set to 1.0 at all elevations*/
  /**/
  "P   335.0  0 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  1 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  2 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  3 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  4 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  5 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  6 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  7 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  8 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0  9 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 10 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 11 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 12 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 13 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 14 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 15 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 16 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 17 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 18 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 19 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 20 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 21 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 22 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 23 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 24 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 25 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 26 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 27 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 28 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "P   335.0 29 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  /**/
  /*-----------------------------------------------------------------------*/
  /**/
  /* L-band. all set to 1.0 at all elevations*/
  /**/
  "L  1500.0  0 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  1 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  2 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  3 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  4 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  5 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  6 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  7 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  8 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0  9 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 10 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 11 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 12 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 13 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 14 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 15 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 16 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 17 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 18 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 19 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 20 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 21 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 22 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 23 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 24 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 25 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 26 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 27 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 28 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "L  1500.0 29 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  /**/
  /*-----------------------------------------------------------------------*/
  /**/
  /* S-band. all set to 1.0 at all elevations*/
  /**/
  "S  3000.0  0 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  1 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  2 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  3 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  4 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  5 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  6 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  7 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  8 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0  9 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 10 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 11 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 12 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 13 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 14 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 15 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 16 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 17 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 18 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 19 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 20 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 21 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 22 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 23 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 24 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 25 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 26 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 27 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 28 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "S  3000.0 29 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  /**/
  /*-----------------------------------------------------------------------*/
  /**/
  /* C-band. all set to 1.0 at all elevations*/
  /**/
  "C  6000.0  0 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  1 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  2 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  3 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  4 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  5 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  6 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  7 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  8 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0  9 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 10 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 11 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 12 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 13 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 14 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 15 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 16 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 17 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 18 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 19 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 20 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 21 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 22 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 23 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 24 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 25 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 26 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 27 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 28 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  "C  6000.0 29 20100101 21000101  1.0000E+00  0.0000E+00  0.0000E+00  0.0000E+00 ",
  /**/
  /* X-band (10000 MHz) gains from flux density run in January 2010.*/
  /**/
  "X 10000.0  2 20100101 20110930  1.0008E+00  5.6372E-06 -6.2722E-07  0.0000E+00 ",
  "X 10000.0  3 20100101 20110930  1.0082E+00 -1.7263E-04 -7.9888E-07  0.0000E+00 ",
  "X 10000.0  4 20100101 20110930  9.9785E-01  2.3347E-04 -4.4907E-06  0.0000E+00 ",
  "X 10000.0  5 20100101 20110930  1.0059E+00 -1.1919E-04 -7.2789E-07  0.0000E+00 ",
  "X 10000.0  6 20100101 20110930  9.8304E-01  6.2248E-04 -4.9611E-06  0.0000E+00 ",
  "X 10000.0  7 20100101 20110930  9.9083E-01  3.9163E-04 -4.0617E-06  0.0000E+00 ",
  "X 10000.0  8 20100101 20110930  9.9530E-01  2.1955E-04 -2.5494E-06  0.0000E+00 ",
  "X 10000.0  9 20100101 20110930  9.9648E-01  1.8410E-04 -2.4030E-06  0.0000E+00 ",
  "X 10000.0 10 20100101 20110930  1.0061E+00 -1.2491E-04 -6.9652E-07  0.0000E+00 ",
  "X 10000.0 13 20100101 20110930  9.8534E-01  5.5566E-04 -4.7283E-06  0.0000E+00 ",
  "X 10000.0 14 20100101 20110930  9.9334E-01  2.9719E-04 -3.2643E-06  0.0000E+00 ",
  "X 10000.0 15 20100101 20110930  1.0034E+00 -1.4938E-04  1.5930E-06  0.0000E+00 ",
  "X 10000.0 16 20100101 20110930  1.0022E+00  1.8983E-05 -1.8499E-06  0.0000E+00 ",
  "X 10000.0 18 20100101 20110930  9.9592E-01  1.4053E-04 -9.6149E-07  0.0000E+00 ",
  "X 10000.0 19 20100101 20110930  1.0044E+00 -1.5068E-04  1.0409E-06  0.0000E+00 ",
  "X 10000.0 20 20100101 20110930  1.0004E+00  5.0020E-05 -1.4954E-06  0.0000E+00 ",
  "X 10000.0 21 20100101 20110930  1.0008E+00  2.9302E-05 -1.2017E-06  0.0000E+00 ",
  "X 10000.0 23 20100101 20110930  1.0064E+00 -2.0699E-04  1.1467E-06  0.0000E+00 ",
  "X 10000.0 24 20100101 20110930  1.0012E+00  2.4954E-05 -1.3584E-06  0.0000E+00 ",
  "X 10000.0 25 20100101 20110930  9.9885E-01  1.3084E-04 -2.5545E-06  0.0000E+00 ",
  "X 10000.0 27 20100101 20110930  9.9300E-01  5.2553E-04 -8.7634E-06  0.0000E+00 ",
  "X 10000.0 28 20100101 20110930  1.0034E+00 -4.3663E-05 -1.0334E-06  0.0000E+00 ",
  "X 10000.0  0 20100101 20110930  9.9878E-01  1.1193E-04 -2.0339E-06  0.0000E+00 ",
  /**/
  /* X-band (10000 MHz) gains from an RPerley run in October 2011.*/
  /**/
  "X 10000.0  1 20111001 20111215  9.9489E-01  2.0382E-04 -1.9027E-06  0.0000E+00 ",
  "X 10000.0  2 20111001 20111215  9.9946E-01  1.2114E-04 -2.6882E-06  0.0000E+00 ",
  "X 10000.0  3 20111001 20111215  9.8744E-01  4.7155E-04 -3.9402E-06  0.0000E+00 ",
  "X 10000.0  4 20111001 20111215  9.8750E-01  4.7705E-04 -4.1124E-06  0.0000E+00 ",
  "X 10000.0  5 20111001 20111215  9.8538E-01 -2.1078E-04  1.4406E-05  0.0000E+00 ",
  "X 10000.0  6 20111001 20111215  9.9327E-01  2.2401E-04 -1.3924E-06  0.0000E+00 ",
  "X 10000.0  7 20111001 20111215  9.8453E-01  5.1819E-04 -3.2884E-06  0.0000E+00 ",
  "X 10000.0  8 20111001 20111215  9.9288E-01  2.8862E-04 -2.7657E-06  0.0000E+00 ",
  "X 10000.0  9 20111001 20111215  9.9375E-01  3.2364E-04 -4.1845E-06  0.0000E+00 ",
  "X 10000.0 10 20111001 20111215  9.8370E-01  5.9531E-04 -4.6974E-06  0.0000E+00 ",
  "X 10000.0 11 20111001 20111215  9.9768E-01  1.3374E-04 -1.8962E-06  0.0000E+00 ",
  "X 10000.0 12 20111001 20111215  9.8687E-01  4.3441E-04 -2.6569E-06  0.0000E+00 ",
  "X 10000.0 13 20111001 20111215  9.8450E-01  6.5557E-04 -6.7025E-06  0.0000E+00 ",
  "X 10000.0 14 20111001 20111215  9.8190E-01  7.7424E-04 -8.0405E-06  0.0000E+00 ",
  "X 10000.0 15 20111001 20111215  9.9164E-01  4.7893E-04 -6.7489E-06  0.0000E+00 ",
  "X 10000.0 16 20111001 20111215  9.9271E-01  8.3837E-05  2.4630E-06  0.0000E+00 ",
  "X 10000.0 17 20111001 20111215  9.8771E-01  5.3615E-04 -5.7255E-06  0.0000E+00 ",
  "X 10000.0 18 20111001 20111215  1.0001E+00  8.9316E-05 -2.2645E-06  0.0000E+00 ",
  "X 10000.0 21 20111001 20111215  9.8954E-01  5.0197E-04 -6.0093E-06  0.0000E+00 ",
  "X 10000.0 22 20111001 20111215  9.8514E-01  5.1973E-04 -3.7071E-06  0.0000E+00 ",
  "X 10000.0 23 20111001 20111215  9.9047E-01  3.8625E-04 -3.6998E-06  0.0000E+00 ",
  "X 10000.0 24 20111001 20111215  9.9029E-01  4.3035E-04 -4.6877E-06  0.0000E+00 ",
  "X 10000.0 25 20111001 20111215  9.8550E-01  6.7932E-04 -7.9195E-06  0.0000E+00 ",
  "X 10000.0 26 20111001 20111215  7.6402E-01  8.7886E-03 -7.2227E-05  0.0000E+00 ",
  "X 10000.0 27 20111001 20111215  9.9484E-01  1.8414E-04 -1.3800E-06  0.0000E+00 ",
  "X 10000.0 28 20111001 20111215  9.9409E-01  2.7613E-04 -3.2113E-06  0.0000E+00 ",
  "X 10000.0  0 20111001 20111215  9.8152E-01  6.9124E-04 -5.7323E-06  0.0000E+00 ",
  /**/
  /* X-band (10000 MHz) gains from flux density run in January 2012.*/
  /**/
  "X 10000.0  1 20111216 20120625  1.0047E+00 -5.8691E-05 -1.4759E-06  0.0000E+00 ",
  "X 10000.0  2 20111216 20120625  1.0037E+00  1.1686E-04 -5.2479E-06  0.0000E+00 ",
  "X 10000.0  3 20111216 20120625  9.8013E-01  7.6227E-04 -6.6358E-06  0.0000E+00 ",
  "X 10000.0  4 20111216 20120625  9.8029E-01  7.1602E-04 -5.5823E-06  0.0000E+00 ",
  "X 10000.0  6 20111216 20120625  9.9668E-01  3.0439E-04 -5.5320E-06  0.0000E+00 ",
  "X 10000.0  7 20111216 20120625  9.9010E-01  4.1670E-04 -4.2278E-06  0.0000E+00 ",
  "X 10000.0  8 20111216 20120625  9.9064E-01  4.6397E-04 -5.7475E-06  0.0000E+00 ",
  "X 10000.0  9 20111216 20120625  9.9688E-01  2.6097E-04 -4.5766E-06  0.0000E+00 ",
  "X 10000.0 10 20111216 20120625  9.8064E-01  7.3810E-04 -6.3507E-06  0.0000E+00 ",
  "X 10000.0 12 20111216 20120625  9.7694E-01  8.7549E-04 -7.4730E-06  0.0000E+00 ",
  "X 10000.0 13 20111216 20120625  9.9495E-01  3.1671E-04 -4.7616E-06  0.0000E+00 ",
  "X 10000.0 14 20111216 20120625  9.9044E-01  4.8002E-04 -6.0250E-06  0.0000E+00 ",
  "X 10000.0 15 20111216 20120625  1.0047E+00  1.4684E-04 -6.5804E-06  0.0000E+00 ",
  "X 10000.0 16 20111216 20120625  9.8262E-01  5.5439E-04 -3.0002E-06  0.0000E+00 ",
  "X 10000.0 17 20111216 20120625  9.7431E-01  1.0710E-03 -1.0722E-05  0.0000E+00 ",
  "X 10000.0 18 20111216 20120625  1.0142E+00 -2.5847E-04 -2.3824E-06  0.0000E+00 ",
  "X 10000.0 19 20111216 20120625  9.9202E-01  4.1134E-04 -5.2938E-06  0.0000E+00 ",
  "X 10000.0 20 20111216 20120625  9.9204E-01  3.8314E-04 -4.6038E-06  0.0000E+00 ",
  "X 10000.0 21 20111216 20120625  9.9829E-01  2.6992E-04 -5.6775E-06  0.0000E+00 ",
  "X 10000.0 22 20111216 20120625  9.8897E-01  4.5611E-04 -4.5096E-06  0.0000E+00 ",
  "X 10000.0 23 20111216 20120625  9.8787E-01  5.2493E-04 -5.5400E-06  0.0000E+00 ",
  "X 10000.0 24 20111216 20120625  9.9019E-01  4.5011E-04 -5.1199E-06  0.0000E+00 ",
  "X 10000.0 25 20111216 20120625  9.7809E-01  8.9664E-04 -8.7222E-06  0.0000E+00 ",
  "X 10000.0 26 20111216 20120625  9.9762E-01  2.2957E-04 -4.2548E-06  0.0000E+00 ",
  "X 10000.0 27 20111216 20120625  9.8467E-01  7.3173E-04 -8.7147E-06  0.0000E+00 ",
  "X 10000.0 28 20111216 20120625  9.9037E-01  4.8801E-04 -6.1833E-06  0.0000E+00 ",
  "X 10000.0  0 20111216 20120625  9.9085E-01  4.5185E-04 -5.5746E-06  0.0000E+00 ",
  /**/
  /* X-band (10000 MHz) gains from gain curve run in September 2012.*/
  /**/
  "X 10000.0  1 20120626 21000101  9.9015E-01  3.4736E-04 -2.5304E-06  0.0000E+00 ",
  "X 10000.0  3 20120626 21000101  9.9306E-01  2.0074E-04 -6.8063E-07  0.0000E+00 ",
  "X 10000.0  4 20120626 21000101  9.8288E-01  5.5746E-04 -3.2374E-06  0.0000E+00 ",
  "X 10000.0  6 20120626 21000101  9.9798E-01  1.0215E-04 -1.2917E-06  0.0000E+00 ",
  "X 10000.0  7 20120626 21000101  9.8410E-01  4.9234E-04 -2.3685E-06  0.0000E+00 ",
  "X 10000.0  8 20120626 21000101  9.9846E-01  1.1753E-04 -1.9764E-06  0.0000E+00 ",
  "X 10000.0 10 20120626 21000101  9.8564E-01  4.0408E-04 -1.1269E-06  0.0000E+00 ",
  "X 10000.0 11 20120626 21000101  9.9472E-01  1.6247E-04 -7.6209E-07  0.0000E+00 ",
  "X 10000.0 12 20120626 21000101  9.8715E-01  4.8300E-04 -4.0452E-06  0.0000E+00 ",
  "X 10000.0 13 20120626 21000101  9.9000E-01  4.0773E-04 -3.9456E-06  0.0000E+00 ",
  "X 10000.0 14 20120626 21000101  9.8445E-01  6.3850E-04 -6.2408E-06  0.0000E+00 ",
  "X 10000.0 15 20120626 21000101  9.9480E-01  3.0207E-04 -4.3008E-06  0.0000E+00 ",
  "X 10000.0 16 20120626 21000101  9.8444E-01  4.3281E-04 -1.0967E-06  0.0000E+00 ",
  "X 10000.0 17 20120626 21000101  9.8970E-01  3.2957E-04 -1.8036E-06  0.0000E+00 ",
  "X 10000.0 18 20120626 21000101  1.0028E+00  7.1804E-05 -3.5409E-06  0.0000E+00 ",
  "X 10000.0 19 20120626 21000101  9.8869E-01  4.7476E-04 -4.7981E-06  0.0000E+00 ",
  "X 10000.0 21 20120626 21000101  9.9347E-01  2.2625E-04 -1.5766E-06  0.0000E+00 ",
  "X 10000.0 22 20120626 21000101  9.8818E-01  4.6881E-04 -4.3355E-06  0.0000E+00 ",
  "X 10000.0 23 20120626 21000101  9.9093E-01  2.9934E-04 -1.8133E-06  0.0000E+00 ",
  "X 10000.0 24 20120626 21000101  9.9485E-01  1.9130E-04 -1.5652E-06  0.0000E+00 ",
  "X 10000.0 25 20120626 21000101  9.9066E-01  4.5914E-04 -5.6440E-06  0.0000E+00 ",
  "X 10000.0 26 20120626 21000101  9.8275E-01  5.6106E-04 -3.2445E-06  0.0000E+00 ",
  "X 10000.0 27 20120626 21000101  1.0001E+00 -3.5399E-04  8.7861E-06  0.0000E+00 ",
  "X 10000.0 28 20120626 21000101  9.8537E-01  5.5943E-04 -4.8439E-06  0.0000E+00 ",
  "X 10000.0  0 20120626 21000101  9.9064E-01  3.3065E-04 -2.4159E-06  0.0000E+00 ",
  /**/
  /* U-band (15000 MHz) gains from an RPerley run in October 2011.*/
  /**/
  /* because this is the first EVLA Ku-band solution, i extend it backwards.*/
  /**/
  "U 15000.0  5 20100101 20111215  9.6929E-01  1.5125E-03 -1.8616E-05  0.0000E+00 ",
  "U 15000.0  6 20100101 20111215  9.9967E-01  3.8812E-04 -9.4985E-06  0.0000E+00 ",
  "U 15000.0  7 20100101 20111215  9.6045E-01  1.8015E-03 -2.0316E-05  0.0000E+00 ",
  "U 15000.0 10 20100101 20111215  9.6668E-01  1.5041E-03 -1.6777E-05  0.0000E+00 ",
  "U 15000.0 12 20100101 20111215  9.4957E-01  2.0338E-03 -1.9326E-05  0.0000E+00 ",
  "U 15000.0 13 20100101 20111215  9.8065E-01  1.1854E-03 -1.7540E-05  0.0000E+00 ",
  "U 15000.0 14 20100101 20111215  9.7709E-01  1.3921E-03 -2.0484E-05  0.0000E+00 ",
  "U 15000.0 16 20100101 20111215  9.7157E-01  1.3093E-03 -1.4961E-05  0.0000E+00 ",
  "U 15000.0 18 20100101 20111215  1.0336E+00 -5.5226E-04 -7.1720E-06  0.0000E+00 ",
  "U 15000.0 21 20100101 20111215  9.9455E-01  5.8860E-04 -1.1311E-05  0.0000E+00 ",
  "U 15000.0 22 20100101 20111215  9.8347E-01  1.0004E-03 -1.4681E-05  0.0000E+00 ",
  "U 15000.0 23 20100101 20111215  9.9599E-01  5.2351E-04 -1.0583E-05  0.0000E+00 ",
  "U 15000.0 24 20100101 20111215  1.0089E+00  2.9051E-04 -1.2798E-05  0.0000E+00 ",
  "U 15000.0 25 20100101 20111215  1.0091E+00  3.9232E-04 -1.5512E-05  0.0000E+00 ",
  "U 15000.0 26 20100101 20111215  1.0120E+00  1.6303E-05 -7.9016E-06  0.0000E+00 ",
  "U 15000.0  0 20100101 20111215  9.8750E-01  8.9243E-04 -1.4498E-05  0.0000E+00 ",
  /**/
  /* U-band (15000 MHz) gains from flux density run in January 2012.*/
  /**/
  "U 15000.0  5 20111216 20120625  9.5104E-01  2.0861E-03 -2.1553E-05  0.0000E+00 ",
  "U 15000.0  6 20111216 20120625  9.8821E-01  9.0221E-04 -1.5189E-05  0.0000E+00 ",
  "U 15000.0  7 20111216 20120625  9.6266E-01  1.5402E-03 -1.5166E-05  0.0000E+00 ",
  "U 15000.0 10 20111216 20120625  9.7758E-01  1.1180E-03 -1.3937E-05  0.0000E+00 ",
  "U 15000.0 12 20111216 20120625  9.6086E-01  1.6494E-03 -1.6770E-05  0.0000E+00 ",
  "U 15000.0 13 20111216 20120625  9.9716E-01  5.3776E-04 -1.1670E-05  0.0000E+00 ",
  "U 15000.0 14 20111216 20120625  9.7520E-01  1.2719E-03 -1.6299E-05  0.0000E+00 ",
  "U 15000.0 16 20111216 20120625  9.7544E-01  1.2514E-03 -1.5937E-05  0.0000E+00 ",
  "U 15000.0 17 20111216 20120625  9.8610E-01  8.9154E-04 -1.3604E-05  0.0000E+00 ",
  "U 15000.0 18 20111216 20120625  1.0526E+00 -1.0200E-03 -7.3957E-06  0.0000E+00 ",
  "U 15000.0 19 20111216 20120625  9.6504E-01  1.7115E-03 -2.0940E-05  0.0000E+00 ",
  "U 15000.0 20 20111216 20120625  9.9736E-01  7.4519E-04 -1.6977E-05  0.0000E+00 ",
  "U 15000.0 21 20111216 20120625  1.0279E+00 -2.9645E-04 -1.0044E-05  0.0000E+00 ",
  "U 15000.0 22 20111216 20120625  1.0350E+00 -9.2027E-04  1.1202E-06  0.0000E+00 ",
  "U 15000.0 23 20111216 20120625  9.8953E-01  1.0372E-03 -1.9389E-05  0.0000E+00 ",
  "U 15000.0 24 20111216 20120625  1.0070E+00  3.4581E-04 -1.3033E-05  0.0000E+00 ",
  "U 15000.0 25 20111216 20120625  9.5055E-01  2.0973E-03 -2.1528E-05  0.0000E+00 ",
  "U 15000.0 26 20111216 20120625  1.0098E+00  2.2398E-04 -1.1722E-05  0.0000E+00 ",
  "U 15000.0  0 20111216 20120625  9.8940E-01  8.4295E-04 -1.4446E-05  0.0000E+00 ",
  /**/
  /* U-band (15000 MHz) gains from gain curve run in September 2012.*/
  /**/
  "U 15000.0  1 20120626 21000101  1.0067E+00  3.5539E-04 -1.3086E-05  0.0000E+00 ",
  "U 15000.0  2 20120626 21000101  1.0134E+00 -6.4292E-05 -6.7936E-06  0.0000E+00 ",
  "U 15000.0  4 20120626 21000101  9.6441E-01  1.5507E-03 -1.6524E-05  0.0000E+00 ",
  "U 15000.0  6 20120626 21000101  1.0043E+00  2.3072E-04 -8.4396E-06  0.0000E+00 ",
  "U 15000.0  7 20120626 21000101  9.6115E-01  1.8682E-03 -2.2427E-05  0.0000E+00 ",
  "U 15000.0 10 20120626 21000101  9.6533E-01  1.3850E-03 -1.2956E-05  0.0000E+00 ",
  "U 15000.0 11 20120626 21000101  9.9451E-01  5.5245E-04 -1.0378E-05  0.0000E+00 ",
  "U 15000.0 12 20120626 21000101  9.7678E-01  8.8538E-04 -7.6228E-06  0.0000E+00 ",
  "U 15000.0 13 20120626 21000101  9.9012E-01  7.5807E-04 -1.2778E-05  0.0000E+00 ",
  "U 15000.0 14 20120626 21000101  9.8207E-01  1.0242E-03 -1.4399E-05  0.0000E+00 ",
  "U 15000.0 16 20120626 21000101  9.7574E-01  1.1078E-03 -1.2529E-05  0.0000E+00 ",
  "U 15000.0 17 20120626 21000101  9.8158E-01  1.1622E-03 -1.7545E-05  0.0000E+00 ",
  "U 15000.0 18 20120626 21000101  1.0391E+00 -6.5316E-04 -8.1174E-06  0.0000E+00 ",
  "U 15000.0 19 20120626 21000101  9.8316E-01  9.4869E-04 -1.3195E-05  0.0000E+00 ",
  "U 15000.0 21 20120626 21000101  1.0070E+00  6.9206E-05 -6.1226E-06  0.0000E+00 ",
  "U 15000.0 22 20120626 21000101  9.9164E-01  7.1537E-04 -1.2660E-05  0.0000E+00 ",
  "U 15000.0 23 20120626 21000101  9.9864E-01  3.2483E-04 -7.2707E-06  0.0000E+00 ",
  "U 15000.0 24 20120626 21000101  1.0117E+00 -7.5490E-05 -5.4129E-06  0.0000E+00 ",
  "U 15000.0 25 20120626 21000101  9.5775E-01  1.7979E-03 -1.8544E-05  0.0000E+00 ",
  "U 15000.0 26 20120626 21000101  9.9231E-01  5.8996E-04 -9.9453E-06  0.0000E+00 ",
  "U 15000.0 27 20120626 21000101  9.7524E-01  1.1403E-03 -1.3031E-05  0.0000E+00 ",
  "U 15000.0  0 20120626 21000101  9.8918E-01  7.4636E-04 -1.1894E-05  0.0000E+00 ",
  /**/
  /* K-band (22000 MHz) gains from flux density run in January 2010.*/
  /**/
  "K 22000.0  1 20100101 20110930  9.6202E-01  1.5520E-03 -1.5063E-05  0.0000E+00 ",
  "K 22000.0  2 20100101 20110930  9.4434E-01  1.8277E-03 -1.0906E-05  0.0000E+00 ",
  "K 22000.0  3 20100101 20110930  9.4102E-01  2.5692E-03 -2.7370E-05  0.0000E+00 ",
  "K 22000.0  4 20100101 20110930  9.7543E-01  1.3520E-03 -1.8442E-05  0.0000E+00 ",
  "K 22000.0  5 20100101 20110930  9.6105E-01  1.6082E-03 -1.5858E-05  0.0000E+00 ",
  "K 22000.0  6 20100101 20110930  9.3959E-01  2.2220E-03 -1.7792E-05  0.0000E+00 ",
  "K 22000.0  7 20100101 20110930  9.5579E-01  2.0426E-03 -2.3432E-05  0.0000E+00 ",
  "K 22000.0  8 20100101 20110930  9.9426E-01  6.3457E-05  2.0012E-06  0.0000E+00 ",
  "K 22000.0  9 20100101 20110930  9.0837E-01  3.6001E-03 -3.2731E-05  0.0000E+00 ",
  "K 22000.0 13 20100101 20110930  9.3629E-01  2.3720E-03 -1.9482E-05  0.0000E+00 ",
  "K 22000.0 14 20100101 20110930  9.4644E-01  2.2481E-03 -2.2727E-05  0.0000E+00 ",
  "K 22000.0 15 20100101 20110930  9.3388E-01  2.2092E-03 -1.3904E-05  0.0000E+00 ",
  "K 22000.0 16 20100101 20110930  9.5809E-01  2.1911E-03 -2.8583E-05  0.0000E+00 ",
  "K 22000.0 18 20100101 20110930  8.8575E-01  3.7930E-03 -2.3417E-05  0.0000E+00 ",
  "K 22000.0 19 20100101 20110930  9.7184E-01  1.2926E-03 -1.4716E-05  0.0000E+00 ",
  "K 22000.0 20 20100101 20110930  9.2769E-01  2.6012E-03 -1.9834E-05  0.0000E+00 ",
  "K 22000.0 21 20100101 20110930  9.0029E-01  3.7142E-03 -3.0535E-05  0.0000E+00 ",
  "K 22000.0 23 20100101 20110930  9.2784E-01  2.5010E-03 -1.7425E-05  0.0000E+00 ",
  "K 22000.0 24 20100101 20110930  9.4039E-01  2.1296E-03 -1.5986E-05  0.0000E+00 ",
  "K 22000.0 25 20100101 20110930  9.6437E-01  1.6798E-03 -1.9725E-05  0.0000E+00 ",
  "K 22000.0 27 20100101 20110930  9.8433E-01  1.1482E-03 -1.8911E-05  0.0000E+00 ",
  "K 22000.0 28 20100101 20110930  9.7131E-01  1.2765E-03 -1.3979E-05  0.0000E+00 ",
  "K 22000.0  0 20100101 20110930  9.4683E-01  2.0906E-03 -1.9037E-05  0.0000E+00 ",
  /**/
  /* K-band (22000 MHz) gains from an RPerley run in October 2011.*/
  /**/
  "K 22000.0  1 20111001 20111215  9.8562E-01  1.1465E-03 -1.9673E-05  0.0000E+00 ",
  "K 22000.0  2 20111001 20111215  1.0084E+00  5.4050E-04 -1.8782E-05  0.0000E+00 ",
  "K 22000.0  3 20111001 20111215  9.5208E-01  2.2561E-03 -2.6453E-05  0.0000E+00 ",
  "K 22000.0  4 20111001 20111215  9.5188E-01  2.1414E-03 -2.3462E-05  0.0000E+00 ",
  "K 22000.0  5 20111001 20111215  9.8371E-01  1.1504E-03 -1.8580E-05  0.0000E+00 ",
  "K 22000.0  6 20111001 20111215  9.9456E-01  7.9888E-04 -1.6571E-05  0.0000E+00 ",
  "K 22000.0  7 20111001 20111215  9.4883E-01  2.0834E-03 -2.0102E-05  0.0000E+00 ",
  "K 22000.0  8 20111001 20111215  9.9569E-01  7.3557E-04 -1.5695E-05  0.0000E+00 ",
  "K 22000.0  9 20111001 20111215  9.7533E-01  1.3011E-03 -1.7111E-05  0.0000E+00 ",
  "K 22000.0 10 20111001 20111215  9.9689E-01  4.3265E-04 -8.8711E-06  0.0000E+00 ",
  "K 22000.0 11 20111001 20111215  9.9120E-01  1.1487E-03 -2.3219E-05  0.0000E+00 ",
  "K 22000.0 12 20111001 20111215  9.6118E-01  1.4750E-03 -1.2612E-05  0.0000E+00 ",
  "K 22000.0 13 20111001 20111215  9.8363E-01  1.2706E-03 -2.1534E-05  0.0000E+00 ",
  "K 22000.0 14 20111001 20111215  9.6350E-01  1.9452E-03 -2.5816E-05  0.0000E+00 ",
  "K 22000.0 15 20111001 20111215  1.0198E+00  4.6113E-04 -2.3918E-05  0.0000E+00 ",
  "K 22000.0 16 20111001 20111215  9.2937E-01  2.7844E-03 -2.5465E-05  0.0000E+00 ",
  "K 22000.0 17 20111001 20111215  9.6947E-01  2.0591E-03 -3.2393E-05  0.0000E+00 ",
  "K 22000.0 18 20111001 20111215  1.0390E+00 -4.6892E-04 -1.2676E-05  0.0000E+00 ",
  "K 22000.0 21 20111001 20111215  9.8552E-01  1.0545E-03 -1.7312E-05  0.0000E+00 ",
  "K 22000.0 22 20111001 20111215  9.7017E-01  1.6571E-03 -2.2782E-05  0.0000E+00 ",
  "K 22000.0 23 20111001 20111215  1.0122E+00 -7.3252E-05 -5.8180E-06  0.0000E+00 ",
  "K 22000.0 24 20111001 20111215  1.0216E+00  2.8977E-05 -1.4204E-05  0.0000E+00 ",
  "K 22000.0 25 20111001 20111215  9.5199E-01  2.6013E-03 -3.5024E-05  0.0000E+00 ",
  "K 22000.0 26 20111001 20111215  1.0480E+00 -9.1977E-04 -7.0338E-06  0.0000E+00 ",
  "K 22000.0 27 20111001 20111215  9.3748E-01  2.6500E-03 -2.7175E-05  0.0000E+00 ",
  "K 22000.0 28 20111001 20111215  9.6657E-01  1.5467E-03 -1.7776E-05  0.0000E+00 ",
  "K 22000.0  0 20111001 20111215  9.8245E-01  1.2234E-03 -1.9617E-05  0.0000E+00 ",
  /**/
  /* K-band (22000 MHz) gains from flux density run in January 2012.*/
  /**/
  "K 22000.0  1 20111216 20120625  1.0575E+00 -1.1512E-03 -7.1645E-06  0.0000E+00 ",
  "K 22000.0  2 20111216 20120625  1.0688E+00 -1.0896E-03 -1.5781E-05  0.0000E+00 ",
  "K 22000.0  3 20111216 20120625  1.0750E+00 -1.7177E-03 -3.9151E-06  0.0000E+00 ",
  "K 22000.0  4 20111216 20120625  1.0427E+00 -7.8463E-04 -7.0729E-06  0.0000E+00 ",
  "K 22000.0  5 20111216 20120625  1.0093E+00  8.8914E-04 -2.8031E-05  0.0000E+00 ",
  "K 22000.0  6 20111216 20120625  9.9201E-01  1.2033E-03 -2.5090E-05  0.0000E+00 ",
  "K 22000.0  7 20111216 20120625  9.7945E-01  1.2882E-03 -1.9363E-05  0.0000E+00 ",
  "K 22000.0  8 20111216 20120625  1.0505E+00 -6.3792E-04 -1.5628E-05  0.0000E+00 ",
  "K 22000.0  9 20111216 20120625  1.0199E+00  2.2734E-04 -1.8145E-05  0.0000E+00 ",
  "K 22000.0 10 20111216 20120625  9.7681E-01  1.7027E-03 -2.8072E-05  0.0000E+00 ",
  "K 22000.0 12 20111216 20120625  9.6099E-01  1.9443E-03 -2.4226E-05  0.0000E+00 ",
  "K 22000.0 13 20111216 20120625  1.0723E+00 -1.7044E-03 -2.6057E-06  0.0000E+00 ",
  "K 22000.0 14 20111216 20120625  1.0260E+00 -4.5890E-05 -1.5121E-05  0.0000E+00 ",
  "K 22000.0 15 20111216 20120625  1.1267E+00 -2.9251E-03 -6.0702E-06  0.0000E+00 ",
  "K 22000.0 16 20111216 20120625  9.9743E-01  1.0894E-03 -2.5631E-05  0.0000E+00 ",
  "K 22000.0 17 20111216 20120625  1.0608E+00 -7.7443E-04 -1.8611E-05  0.0000E+00 ",
  "K 22000.0 18 20111216 20120625  1.1145E+00 -2.7621E-03 -2.5116E-06  0.0000E+00 ",
  "K 22000.0 19 20111216 20120625  9.8657E-01  1.2006E-03 -2.1622E-05  0.0000E+00 ",
  "K 22000.0 20 20111216 20120625  1.0509E+00 -5.4842E-04 -1.8103E-05  0.0000E+00 ",
  "K 22000.0 21 20111216 20120625  1.0422E+00 -5.5806E-04 -1.2450E-05  0.0000E+00 ",
  "K 22000.0 22 20111216 20120625  1.1033E+00 -3.2601E-03  1.6925E-05  0.0000E+00 ",
  "K 22000.0 23 20111216 20120625  1.0653E+00 -8.7743E-04 -1.8897E-05  0.0000E+00 ",
  "K 22000.0 24 20111216 20120625  1.0278E+00 -1.2685E-04 -1.4193E-05  0.0000E+00 ",
  "K 22000.0 26 20111216 20120625  1.0634E+00 -1.6474E-03  1.5312E-06  0.0000E+00 ",
  "K 22000.0 27 20111216 20120625  1.0059E+00  4.4446E-04 -1.4784E-05  0.0000E+00 ",
  "K 22000.0 28 20111216 20120625  1.0189E+00 -1.1315E-04 -8.9700E-06  0.0000E+00 ",
  "K 22000.0  0 20111216 20120625  1.0383E+00 -4.1293E-04 -1.3599E-05  0.0000E+00 ",
  /**/
  /* K-band (22000 MHz) gains from gain curve run in September 2012.*/
  /**/
  "K 22000.0  1 20120626 21000101  1.0078E+00  1.9643E-04 -9.7548E-06  0.0000E+00 ",
  "K 22000.0  2 20120626 21000101  1.0142E+00  2.0870E-04 -1.4086E-05  0.0000E+00 ",
  "K 22000.0  3 20120626 21000101  9.7026E-01  1.3125E-03 -1.4225E-05  0.0000E+00 ",
  "K 22000.0  4 20120626 21000101  9.6764E-01  1.1308E-03 -8.0430E-06  0.0000E+00 ",
  "K 22000.0  6 20120626 21000101  9.9434E-01  7.5774E-04 -1.5404E-05  0.0000E+00 ",
  "K 22000.0  7 20120626 21000101  9.4415E-01  2.1499E-03 -1.8838E-05  0.0000E+00 ",
  "K 22000.0  8 20120626 21000101  9.8275E-01  7.7302E-04 -8.5464E-06  0.0000E+00 ",
  "K 22000.0 10 20120626 21000101  9.4950E-01  1.6714E-03 -1.0221E-05  0.0000E+00 ",
  "K 22000.0 11 20120626 21000101  9.7408E-01  9.9561E-04 -8.6887E-06  0.0000E+00 ",
  "K 22000.0 12 20120626 21000101  9.6172E-01  1.3214E-03 -9.1096E-06  0.0000E+00 ",
  "K 22000.0 13 20120626 21000101  9.8476E-01  1.0428E-03 -1.6543E-05  0.0000E+00 ",
  "K 22000.0 14 20120626 21000101  9.8011E-01  1.2532E-03 -1.8896E-05  0.0000E+00 ",
  "K 22000.0 15 20120626 21000101  1.0299E+00  2.3498E-05 -1.9256E-05  0.0000E+00 ",
  "K 22000.0 16 20120626 21000101  9.5465E-01  1.6025E-03 -1.1720E-05  0.0000E+00 ",
  "K 22000.0 17 20120626 21000101  9.8257E-01  1.1259E-03 -1.7252E-05  0.0000E+00 ",
  "K 22000.0 18 20120626 21000101  1.0409E+00 -4.9382E-04 -1.3187E-05  0.0000E+00 ",
  "K 22000.0 19 20120626 21000101  9.7195E-01  1.3235E-03 -1.5555E-05  0.0000E+00 ",
  "K 22000.0 21 20120626 21000101  9.6905E-01  1.5891E-03 -2.0385E-05  0.0000E+00 ",
  "K 22000.0 22 20120626 21000101  9.8235E-01  1.2108E-03 -1.9241E-05  0.0000E+00 ",
  "K 22000.0 24 20120626 21000101  9.9930E-01  2.2149E-04 -5.0981E-06  0.0000E+00 ",
  "K 22000.0 26 20120626 21000101  9.8228E-01  8.4332E-04 -1.0009E-05  0.0000E+00 ",
  "K 22000.0 27 20120626 21000101  9.6100E-01  1.6453E-03 -1.6756E-05  0.0000E+00 ",
  "K 22000.0 28 20120626 21000101  9.7385E-01  9.7905E-04 -8.1298E-06  0.0000E+00 ",
  "K 22000.0  0 20120626 21000101  9.8169E-01  9.9500E-04 -1.3432E-05  0.0000E+00 ",
  /**/
  /* A-band (33000 MHz) gains from flux density run in January 2010.*/
  /**/
  "A 33000.0  2 20100101 20110930  8.4905E-01  5.8594E-03 -5.2140E-05  0.0000E+00 ",
  "A 33000.0  3 20100101 20110930  8.4937E-01  5.8871E-03 -5.3036E-05  0.0000E+00 ",
  "A 33000.0  4 20100101 20110930  8.4868E-01  5.8569E-03 -5.1848E-05  0.0000E+00 ",
  "A 33000.0  5 20100101 20110930  8.9372E-01  4.5805E-03 -4.8088E-05  0.0000E+00 ",
  "A 33000.0  6 20100101 20110930  9.1099E-01  3.9482E-03 -4.3076E-05  0.0000E+00 ",
  "A 33000.0  7 20100101 20110930  8.6992E-01  5.5862E-03 -5.8358E-05  0.0000E+00 ",
  "A 33000.0  8 20100101 20110930  8.4346E-01  5.7684E-03 -4.6375E-05  0.0000E+00 ",
  "A 33000.0  9 20100101 20110930  9.5038E-01  2.9920E-03 -4.3788E-05  0.0000E+00 ",
  "A 33000.0 15 20100101 20110930  7.4522E-01  8.3111E-03 -4.8540E-05  0.0000E+00 ",
  "A 33000.0 19 20100101 20110930  9.7343E-01  2.4607E-03 -4.4913E-05  0.0000E+00 ",
  "A 33000.0 20 20100101 20110930  8.4336E-01  6.3371E-03 -6.0526E-05  0.0000E+00 ",
  "A 33000.0 23 20100101 20110930  9.1125E-01  3.6823E-03 -3.6588E-05  0.0000E+00 ",
  "A 33000.0 24 20100101 20110930  8.8359E-01  4.0886E-03 -2.9459E-05  0.0000E+00 ",
  "A 33000.0 25 20100101 20110930  8.7083E-01  5.1225E-03 -4.7330E-05  0.0000E+00 ",
  "A 33000.0 27 20100101 20110930  9.7147E-01  2.0155E-03 -3.2556E-05  0.0000E+00 ",
  "A 33000.0 28 20100101 20110930  9.4768E-01  2.1873E-03 -2.1980E-05  0.0000E+00 ",
  "A 33000.0  0 20100101 20110930  8.8512E-01  4.6684E-03 -4.4913E-05  0.0000E+00 ",
  /**/
  /* A-band (33000 MHz) gains from an RPerley run in October 2011.*/
  /**/
  "A 33000.0  1 20111001 20111215  9.1785E-01  5.0637E-03 -7.5250E-05  0.0000E+00 ",
  "A 33000.0  2 20111001 20111215  9.3347E-01  3.5828E-03 -4.7993E-05  0.0000E+00 ",
  "A 33000.0  3 20111001 20111215  9.2697E-01  3.7121E-03 -4.7160E-05  0.0000E+00 ",
  "A 33000.0  4 20111001 20111215  9.2101E-01  3.8094E-03 -4.5868E-05  0.0000E+00 ",
  "A 33000.0  5 20111001 20111215  9.0446E-01  4.4791E-03 -5.2263E-05  0.0000E+00 ",
  "A 33000.0  6 20111001 20111215  9.5047E-01  2.7964E-03 -3.8951E-05  0.0000E+00 ",
  "A 33000.0  7 20111001 20111215  9.3923E-01  2.7764E-03 -3.1431E-05  0.0000E+00 ",
  "A 33000.0  8 20111001 20111215  9.8553E-01  1.1855E-03 -2.0595E-05  0.0000E+00 ",
  "A 33000.0  9 20111001 20111215  8.8326E-01  4.5728E-03 -4.1357E-05  0.0000E+00 ",
  "A 33000.0  1 20111001 20111215  9.5539E-01  4.4200E-03 -8.2620E-05  0.0000E+00 ",
  "A 33000.0 10 20111001 20111215  9.4362E-01  3.2658E-03 -4.6410E-05  0.0000E+00 ",
  "A 33000.0 11 20111001 20111215  9.7349E-01  2.2725E-03 -4.0243E-05  0.0000E+00 ",
  "A 33000.0 12 20111001 20111215  9.2935E-01  3.2939E-03 -3.8189E-05  0.0000E+00 ",
  "A 33000.0 13 20111001 20111215  9.7404E-01  2.1029E-03 -3.6344E-05  0.0000E+00 ",
  "A 33000.0 14 20111001 20111215  9.4827E-01  3.0886E-03 -4.4885E-05  0.0000E+00 ",
  "A 33000.0 15 20111001 20111215  9.2556E-01  4.1462E-03 -5.7133E-05  0.0000E+00 ",
  "A 33000.0 16 20111001 20111215  9.0361E-01  3.8686E-03 -3.6470E-05  0.0000E+00 ",
  "A 33000.0 17 20111001 20111215  8.4054E-01  7.1000E-03 -7.7835E-05  0.0000E+00 ",
  "A 33000.0 18 20111001 20111215  9.3967E-01  3.2091E-03 -4.2525E-05  0.0000E+00 ",
  "A 33000.0 21 20111001 20111215  9.6804E-01  1.9818E-03 -2.9573E-05  0.0000E+00 ",
  "A 33000.0 22 20111001 20111215  9.3699E-01  3.0633E-03 -3.7202E-05  0.0000E+00 ",
  "A 33000.0 23 20111001 20111215  9.5583E-01  2.2335E-03 -2.8234E-05  0.0000E+00 ",
  "A 33000.0 24 20111001 20111215  1.0049E+00  7.2869E-04 -2.1297E-05  0.0000E+00 ",
  "A 33000.0 25 20111001 20111215  9.1704E-01  4.1232E-03 -5.1229E-05  0.0000E+00 ",
  "A 33000.0 26 20111001 20111215  1.0388E+00  7.3596E-05 -2.6093E-05  0.0000E+00 ",
  "A 33000.0 27 20111001 20111215  9.1409E-01  3.4849E-03 -3.3427E-05  0.0000E+00 ",
  "A 33000.0 28 20111001 20111215  9.2533E-01  2.8005E-03 -2.3341E-05  0.0000E+00 ",
  "A 33000.0  0 20111001 20111215  9.3914E-01  3.2307E-03 -4.2730E-05  0.0000E+00 ",
  /**/
  /* A-band (33000 MHz) gains from flux density run in January 2012.*/
  /**/
  "A 33000.0  2 20111216 20120625  1.0182E+00  1.3992E-03 -4.6374E-05  0.0000E+00 ",
  "A 33000.0  3 20111216 20120625  1.0801E+00 -9.1657E-04 -2.7179E-05  0.0000E+00 ",
  "A 33000.0  4 20111216 20120625  1.0548E+00 -6.2487E-04 -1.8640E-05  0.0000E+00 ",
  "A 33000.0  5 20111216 20120625  9.4829E-01  3.5005E-03 -5.5194E-05  0.0000E+00 ",
  "A 33000.0  6 20111216 20120625  1.0338E+00 -1.1628E-04 -1.8203E-05  0.0000E+00 ",
  "A 33000.0  7 20111216 20120625  1.0212E+00  9.7707E-04 -3.7706E-05  0.0000E+00 ",
  "A 33000.0  8 20111216 20120625  1.0573E+00 -1.3673E-04 -3.2419E-05  0.0000E+00 ",
  "A 33000.0  9 20111216 20120625  1.0166E+00  7.6186E-04 -2.9428E-05  0.0000E+00 ",
  "A 33000.0 10 20111216 20120625  1.0709E+00 -8.0202E-04 -2.4251E-05  0.0000E+00 ",
  "A 33000.0 12 20111216 20120625  1.0339E+00  3.8182E-04 -3.0749E-05  0.0000E+00 ",
  "A 33000.0 13 20111216 20120625  9.7436E-01  2.3024E-03 -4.1537E-05  0.0000E+00 ",
  "A 33000.0 14 20111216 20120625  1.0378E+00  3.6849E-04 -3.2822E-05  0.0000E+00 ",
  "A 33000.0 15 20111216 20120625  1.0424E+00  6.0831E-04 -4.1683E-05  0.0000E+00 ",
  "A 33000.0 16 20111216 20120625  8.9687E-01  4.4233E-03 -4.6124E-05  0.0000E+00 ",
  "A 33000.0 17 20111216 20120625  9.5670E-01  2.7654E-03 -4.2072E-05  0.0000E+00 ",
  "A 33000.0 18 20111216 20120625  1.0465E+00  2.1947E-04 -3.4564E-05  0.0000E+00 ",
  "A 33000.0 19 20111216 20120625  9.9461E-01  8.5998E-04 -1.8132E-05  0.0000E+00 ",
  "A 33000.0 20 20111216 20120625  1.0298E+00  3.1996E-04 -2.6647E-05  0.0000E+00 ",
  "A 33000.0 21 20111216 20120625  1.0810E+00 -1.6906E-03 -8.3885E-06  0.0000E+00 ",
  "A 33000.0 22 20111216 20120625  1.0901E+00 -1.9135E-03 -8.4725E-06  0.0000E+00 ",
  "A 33000.0 23 20111216 20120625  1.0504E+00  1.6318E-05 -3.1903E-05  0.0000E+00 ",
  "A 33000.0 24 20111216 20120625  1.0856E+00 -1.1702E-03 -2.4232E-05  0.0000E+00 ",
  "A 33000.0 25 20111216 20120625  1.0163E+00  1.4443E-04 -1.3782E-05  0.0000E+00 ",
  "A 33000.0 26 20111216 20120625  1.0607E+00 -4.6055E-04 -2.6439E-05  0.0000E+00 ",
  "A 33000.0 27 20111216 20120625  1.0010E+00  1.5163E-03 -3.8525E-05  0.0000E+00 ",
  "A 33000.0 28 20111216 20120625  1.0361E+00 -4.6140E-04 -1.1038E-05  0.0000E+00 ",
  "A 33000.0  0 20111216 20120625  1.0283E+00  4.7194E-04 -2.9478E-05  0.0000E+00 ",
  /**/
  /* A-band (33000 MHz) gains from gain curve run in September 2012.*/
  /**/
  "A 33000.0  1 20120626 21000101  9.7422E-01  1.8107E-03 -2.9151E-05  0.0000E+00 ",
  "A 33000.0  2 20120626 21000101  9.4827E-01  2.7722E-03 -3.6975E-05  0.0000E+00 ",
  "A 33000.0  3 20120626 21000101  9.4279E-01  2.7839E-03 -3.3839E-05  0.0000E+00 ",
  "A 33000.0  6 20120626 21000101  9.5310E-01  2.2866E-03 -2.7853E-05  0.0000E+00 ",
  "A 33000.0  7 20120626 21000101  9.0336E-01  4.3219E-03 -4.7649E-05  0.0000E+00 ",
  "A 33000.0  8 20120626 21000101  9.7693E-01  1.3142E-03 -1.8439E-05  0.0000E+00 ",
  "A 33000.0 10 20120626 21000101  9.2895E-01  3.3487E-03 -3.9311E-05  0.0000E+00 ",
  "A 33000.0 11 20120626 21000101  9.8766E-01  8.9714E-04 -1.4718E-05  0.0000E+00 ",
  "A 33000.0 12 20120626 21000101  9.8803E-01  1.0365E-03 -1.8435E-05  0.0000E+00 ",
  "A 33000.0 13 20120626 21000101  9.3014E-01  2.8511E-03 -2.7614E-05  0.0000E+00 ",
  "A 33000.0 14 20120626 21000101  9.6765E-01  2.2383E-03 -3.5740E-05  0.0000E+00 ",
  "A 33000.0 15 20120626 21000101  9.7009E-01  1.9226E-03 -2.9375E-05  0.0000E+00 ",
  "A 33000.0 16 20120626 21000101  9.0919E-01  3.5304E-03 -3.1506E-05  0.0000E+00 ",
  "A 33000.0 17 20120626 21000101  8.2436E-01  6.8249E-03 -6.0849E-05  0.0000E+00 ",
  "A 33000.0 18 20120626 21000101  9.3941E-01  3.1545E-03 -4.0995E-05  0.0000E+00 ",
  "A 33000.0 19 20120626 21000101  8.7897E-01  4.6668E-03 -4.1028E-05  0.0000E+00 ",
  "A 33000.0 21 20120626 21000101  9.8912E-01  8.9625E-04 -1.5609E-05  0.0000E+00 ",
  "A 33000.0 22 20120626 21000101  9.7435E-01  1.6933E-03 -2.6300E-05  0.0000E+00 ",
  "A 33000.0 23 20120626 21000101  9.7131E-01  1.5578E-03 -2.1014E-05  0.0000E+00 ",
  "A 33000.0 24 20120626 21000101  9.9296E-01  9.3911E-04 -1.9077E-05  0.0000E+00 ",
  "A 33000.0 25 20120626 21000101  9.4448E-01  2.4124E-03 -2.5609E-05  0.0000E+00 ",
  "A 33000.0 26 20120626 21000101  9.4633E-01  2.8416E-03 -3.7496E-05  0.0000E+00 ",
  "A 33000.0 27 20120626 21000101  9.5250E-01  1.7151E-03 -1.3188E-05  0.0000E+00 ",
  "A 33000.0  0 20120626 21000101  9.4757E-01  2.5138E-03 -3.0077E-05  0.0000E+00 ",
  /**/
  /* Q-band (45000 MHz) gains from flux density run in January 2010.*/
  /**/
  "Q 45000.0  1 20100101 20110930  8.3860E-01  7.7367E-03 -9.2543E-05  0.0000E+00 ",
  "Q 45000.0  2 20100101 20110930  7.2376E-01  1.0116E-02 -8.0258E-05  0.0000E+00 ",
  "Q 45000.0  3 20100101 20110930  8.6613E-01  6.0460E-03 -6.7480E-05  0.0000E+00 ",
  "Q 45000.0  4 20100101 20110930  7.4478E-01  1.0536E-02 -1.0390E-04  0.0000E+00 ",
  "Q 45000.0  5 20100101 20110930  8.0594E-01  8.0564E-03 -8.0121E-05  0.0000E+00 ",
  "Q 45000.0  6 20100101 20110930  7.4534E-01  9.5008E-03 -7.8357E-05  0.0000E+00 ",
  "Q 45000.0  7 20100101 20110930  6.5745E-01  1.3038E-02 -1.1186E-04  0.0000E+00 ",
  "Q 45000.0  8 20100101 20110930  7.6204E-01  8.7599E-03 -7.0272E-05  0.0000E+00 ",
  "Q 45000.0  9 20100101 20110930  8.1295E-01  8.1876E-03 -8.7785E-05  0.0000E+00 ",
  "Q 45000.0 13 20100101 20110930  6.8473E-01  1.0527E-02 -6.6143E-05  0.0000E+00 ",
  "Q 45000.0 14 20100101 20110930  8.5666E-01  6.0407E-03 -6.1428E-05  0.0000E+00 ",
  "Q 45000.0 15 20100101 20110930  6.8187E-01  1.0039E-02 -5.2139E-05  0.0000E+00 ",
  "Q 45000.0 16 20100101 20110930  7.9304E-01  9.5777E-03 -1.1009E-04  0.0000E+00 ",
  "Q 45000.0 18 20100101 20110930  6.1610E-01  1.3494E-02 -9.7425E-05  0.0000E+00 ",
  "Q 45000.0 19 20100101 20110930  8.2189E-01  8.3426E-03 -9.7250E-05  0.0000E+00 ",
  "Q 45000.0 21 20100101 20110930  8.1561E-01  6.3885E-03 -4.4472E-05  0.0000E+00 ",
  "Q 45000.0 23 20100101 20110930  7.2312E-01  1.0182E-02 -8.1489E-05  0.0000E+00 ",
  "Q 45000.0 24 20100101 20110930  7.1158E-01  1.0312E-02 -7.7540E-05  0.0000E+00 ",
  "Q 45000.0 25 20100101 20110930  8.2341E-01  7.6185E-03 -8.0093E-05  0.0000E+00 ",
  "Q 45000.0 27 20100101 20110930  7.1895E-01  9.1476E-03 -5.3033E-05  0.0000E+00 ",
  "Q 45000.0 28 20100101 20110930  7.9940E-01  7.7001E-03 -6.7125E-05  0.0000E+00 ",
  "Q 45000.0  0 20100101 20110930  7.6198E-01  9.1135E-03 -7.9077E-05  0.0000E+00 ",
  /**/
    /* Q-band (45000 MHz) gains from an RPerley run in October 2011.*/
    /**/
    "Q 45000.0  1 20111001 20111215  8.7338E-01  6.9768E-03 -9.5281E-05  0.0000E+00 ",
    "Q 45000.0  2 20111001 20111215  9.5891E-01  4.1841E-03 -7.8921E-05  0.0000E+00 ",
    "Q 45000.0  3 20111001 20111215  8.8097E-01  5.8931E-03 -7.2935E-05  0.0000E+00 ",
    "Q 45000.0  4 20111001 20111215  8.6030E-01  6.6831E-03 -7.9762E-05  0.0000E+00 ",
    "Q 45000.0  5 20111001 20111215  8.8454E-01  6.1293E-03 -8.1072E-05  0.0000E+00 ",
    "Q 45000.0  6 20111001 20111215  9.4104E-01  3.7392E-03 -5.6632E-05  0.0000E+00 ",
    "Q 45000.0  7 20111001 20111215  8.0726E-01  7.6812E-03 -7.1568E-05  0.0000E+00 ",
    "Q 45000.0  8 20111001 20111215  9.6672E-01  2.3728E-03 -3.8523E-05  0.0000E+00 ",
    "Q 45000.0  9 20111001 20111215  8.3890E-01  6.8802E-03 -7.1315E-05  0.0000E+00 ",
    "Q 45000.0 10 20111001 20111215  8.9027E-01  6.1761E-03 -8.5824E-05  0.0000E+00 ",
    "Q 45000.0 11 20111001 20111215  9.9460E-01  3.1562E-03 -7.5530E-05  0.0000E+00 ",
    "Q 45000.0 12 20111001 20111215  8.2033E-01  7.3720E-03 -7.2008E-05  0.0000E+00 ",
    "Q 45000.0 13 20111001 20111215  9.4697E-01  3.6506E-03 -5.8122E-05  0.0000E+00 ",
    "Q 45000.0 14 20111001 20111215  8.6548E-01  6.6461E-03 -8.2074E-05  0.0000E+00 ",
    "Q 45000.0 15 20111001 20111215  1.0895E+00  4.6261E-04 -6.7525E-05  0.0000E+00 ",
    "Q 45000.0 16 20111001 20111215  8.1507E-01  7.6896E-03 -7.6660E-05  0.0000E+00 ",
    "Q 45000.0 17 20111001 20111215  8.1053E-01  8.3679E-03 -9.0777E-05  0.0000E+00 ",
    "Q 45000.0 18 20111001 20111215  1.0049E+00  2.5116E-03 -6.5869E-05  0.0000E+00 ",
    "Q 45000.0 21 20111001 20111215  1.0078E+00  2.3116E-03 -6.2694E-05  0.0000E+00 ",
    "Q 45000.0 22 20111001 20111215  8.5916E-01  6.8307E-03 -8.2741E-05  0.0000E+00 ",
    "Q 45000.0 23 20111001 20111215  1.0085E+00  1.8679E-03 -5.1995E-05  0.0000E+00 ",
    "Q 45000.0 24 20111001 20111215  1.0162E+00  3.4505E-04 -1.8770E-05  0.0000E+00 ",
    "Q 45000.0 25 20111001 20111215  8.4598E-01  7.9118E-03 -1.0153E-04  0.0000E+00 ",
    "Q 45000.0 26 20111001 20111215  1.1134E+00 -2.0395E-03 -1.9918E-05  0.0000E+00 ",
    "Q 45000.0 27 20111001 20111215  7.3140E-01  1.0702E-02 -9.9672E-05  0.0000E+00 ",
    "Q 45000.0 28 20111001 20111215  8.4374E-01  6.6137E-03 -6.7682E-05  0.0000E+00 ",
    "Q 45000.0  0 20111001 20111215  9.1056E-01  5.0441E-03 -7.0204E-05  0.0000E+00 ",
    /**/
    /* Q-band (45000 MHz) gains from flux density run in January 2012.*/
    /**/
    "Q 45000.0  1 20111216 20120625  1.1082E+00 -8.2130E-04 -4.7080E-05  0.0000E+00 ",
    "Q 45000.0  2 20111216 20120625  1.2452E+00 -5.6737E-03 -1.1389E-05  0.0000E+00 ",
    "Q 45000.0  3 20111216 20120625  1.1710E+00 -4.0126E-03 -6.5866E-06  0.0000E+00 ",
    "Q 45000.0  4 20111216 20120625  1.1774E+00 -3.5283E-03 -2.2661E-05  0.0000E+00 ",
    "Q 45000.0  5 20111216 20120625  1.0573E+00  9.5765E-04 -5.9771E-05  0.0000E+00 ",
    "Q 45000.0  6 20111216 20120625  9.4989E-01  4.9292E-03 -9.1912E-05  0.0000E+00 ",
    "Q 45000.0  7 20111216 20120625  1.0251E+00  1.0989E-03 -4.3174E-05  0.0000E+00 ",
    "Q 45000.0  8 20111216 20120625  1.2278E+00 -5.2652E-03 -1.0742E-05  0.0000E+00 ",
    "Q 45000.0  9 20111216 20120625  1.0413E+00  8.6835E-04 -4.7517E-05  0.0000E+00 ",
    "Q 45000.0 10 20111216 20120625  1.0411E+00  1.0920E-03 -5.3007E-05  0.0000E+00 ",
    "Q 45000.0 12 20111216 20120625  1.0262E+00  1.9657E-03 -6.5502E-05  0.0000E+00 ",
    "Q 45000.0 13 20111216 20120625  1.1620E+00 -3.3476E-03 -1.7538E-05  0.0000E+00 ",
    "Q 45000.0 14 20111216 20120625  1.2687E+00 -8.6461E-03  4.8234E-05  0.0000E+00 ",
    "Q 45000.0 15 20111216 20120625  1.4502E+00 -1.1231E-02 -6.4192E-07  0.0000E+00 ",
    "Q 45000.0 16 20111216 20120625  9.6719E-01  3.7080E-03 -7.2193E-05  0.0000E+00 ",
    "Q 45000.0 17 20111216 20120625  1.1130E+00 -1.0970E-03 -4.3206E-05  0.0000E+00 ",
    "Q 45000.0 18 20111216 20120625  1.3111E+00 -7.4939E-03 -7.1084E-06  0.0000E+00 ",
    "Q 45000.0 19 20111216 20120625  1.0074E+00  2.5520E-03 -6.8454E-05  0.0000E+00 ",
    "Q 45000.0 20 20111216 20120625  1.2317E+00 -4.8638E-03 -2.3189E-05  0.0000E+00 ",
    "Q 45000.0 21 20111216 20120625  1.1953E+00 -3.3042E-03 -3.9483E-05  0.0000E+00 ",
    "Q 45000.0 22 20111216 20120625  1.1715E+00 -4.3210E-03  8.5311E-07  0.0000E+00 ",
    "Q 45000.0 23 20111216 20120625  1.2057E+00 -3.0336E-03 -5.2707E-05  0.0000E+00 ",
    "Q 45000.0 24 20111216 20120625  1.0454E+00  1.3935E-03 -6.3209E-05  0.0000E+00 ",
    "Q 45000.0 25 20111216 20120625  1.0997E+00 -1.1111E-03 -3.4546E-05  0.0000E+00 ",
    "Q 45000.0 26 20111216 20120625  1.1981E+00 -4.4238E-03 -1.3200E-05  0.0000E+00 ",
    "Q 45000.0 27 20111216 20120625  9.9374E-01  2.7020E-03 -6.3638E-05  0.0000E+00 ",
    "Q 45000.0 28 20111216 20120625  1.1071E+00 -1.5904E-03 -2.7162E-05  0.0000E+00 ",
    "Q 45000.0  0 20111216 20120625  1.1333E+00 -1.9455E-03 -3.4664E-05  0.0000E+00 ",
    /**/
    /* Q-band (45000 MHz) gains from gain curve run in September 2012.*/
    /**/
    "Q 45000.0  1 20120626 21000101  9.6565E-01  2.8696E-03 -5.0269E-05  0.0000E+00 ",
    "Q 45000.0  2 20120626 21000101  1.0193E+00  6.1995E-04 -2.7582E-05  0.0000E+00 ",
    "Q 45000.0  3 20120626 21000101  8.3930E-01  7.5935E-03 -8.9399E-05  0.0000E+00 ",
    "Q 45000.0  4 20120626 21000101  9.1973E-01  4.2791E-03 -5.6811E-05  0.0000E+00 ",
    "Q 45000.0  6 20120626 21000101  9.2570E-01  4.0572E-03 -5.4991E-05  0.0000E+00 ",
    "Q 45000.0  7 20120626 21000101  8.1253E-01  7.8044E-03 -7.7938E-05  0.0000E+00 ",
    "Q 45000.0 10 20120626 21000101  8.3506E-01  6.8608E-03 -6.8431E-05  0.0000E+00 ",
    "Q 45000.0 11 20120626 21000101  9.8705E-01  1.7004E-03 -3.4414E-05  0.0000E+00 ",
    "Q 45000.0 12 20120626 21000101  8.9882E-01  4.6248E-03 -5.2384E-05  0.0000E+00 ",
    "Q 45000.0 13 20120626 21000101  9.0666E-01  4.5627E-03 -5.5729E-05  0.0000E+00 ",
    "Q 45000.0 14 20120626 21000101  9.5127E-01  3.4793E-03 -5.6523E-05  0.0000E+00 ",
    "Q 45000.0 15 20120626 21000101  1.0862E+00 -2.1943E-04 -4.8385E-05  0.0000E+00 ",
    "Q 45000.0 16 20120626 21000101  8.1581E-01  7.4697E-03 -7.1623E-05  0.0000E+00 ",
    "Q 45000.0 17 20120626 21000101  8.6614E-01  6.9414E-03 -8.9870E-05  0.0000E+00 ",
    "Q 45000.0 18 20120626 21000101  1.0175E+00  1.9786E-03 -6.0415E-05  0.0000E+00 ",
    "Q 45000.0 19 20120626 21000101  8.3520E-01  7.4227E-03 -8.2564E-05  0.0000E+00 ",
    "Q 45000.0 21 20120626 21000101  9.8497E-01  1.9708E-03 -3.9875E-05  0.0000E+00 ",
    "Q 45000.0 22 20120626 21000101  9.5888E-01  3.0099E-03 -4.9547E-05  0.0000E+00 ",
    "Q 45000.0 23 20120626 21000101  1.0100E+00  1.3265E-03 -3.9413E-05  0.0000E+00 ",
    "Q 45000.0 24 20120626 21000101  9.0717E-01  4.7252E-03 -6.0111E-05  0.0000E+00 ",
    "Q 45000.0 25 20120626 21000101  8.7439E-01  5.4337E-03 -5.7336E-05  0.0000E+00 ",
    "Q 45000.0 26 20120626 21000101  9.3125E-01  3.7500E-03 -5.0783E-05  0.0000E+00 ",
    "Q 45000.0 28 20120626 21000101  9.4445E-01  2.8664E-03 -3.6939E-05  0.0000E+00 ",
    "Q 45000.0  0 20120626 21000101  9.2576E-01  4.1364E-03 -5.7013E-05  0.0000E+00 ",
   NULL
  };
  gchar *line, band, datestr[12];
  /* Number of bands tabulated */
  olong nband=10;
  /* Band codes in Table */
  gchar BandCode[] = {'4', 'P', 'L', 'S', 'C', 'X', 'U', 'K', 'A', 'Q'};
  /* Upper frequency of bands in BandCodes */
  odouble tfreq;
  odouble upFreq[] = 
  /*  4        P      L       S      C       X      U        K       A      Q */
    {100.0e6, 900.0e6, 2.0e9, 3.7e9, 7.5e9, 12.0e9, 18.0e9, 26.5e9, 40.0e9, 50.0e9};
  
  /* Lookup band code */
  band = BandCode[0];
    for (i=0; i<nband; i++) {
	   band = BandCode[i];
    if (Freq<upFreq[i]) break;
  }
  
  /* Make temp gain array */
  garray = g_malloc0(8*sizeof(ofloat*));
  for (i=0; i<8; i++) garray[i] = g_malloc0(4*sizeof(ofloat));

  /* Convert JD to data string */
  ObitUVDescJD2Date (JD, datestr);
  /* check */
  if (strncmp(datestr, "BAD DATE", 8)) {
    /* parse date */
    datestr[4] = ' '; datestr[7] = ' ';  /* remove '-' */
    n = sscanf (datestr, "%4d%3d%3d", &yr, &mo, &day);
    
  } else {  /* having a bad day - make it April 1, 2010 */
    yr  = 2010;
    mo  = 4;
    day = 1;
  }
  /* Put together test integer */
  date = yr*10000 + mo*100. + day;

  /*          1         2         3         4         5         6         7         */
  /*01234567890123456789012345678901234567890123456789012345678901234567890123456789*/
  /*Q 48425.0  0 20120101 21000101  7.1791E-03 -4.8166E-04  7.5139E-06  0.0000E+00*/
  /* Loop through table */
  iline   = 0;
  found   = FALSE;
  gotBand = FALSE;
  while (gainLines[iline]) {
    line = gainLines[iline];
    /* Band too far? */
    if ((line[0]!=band) && gotBand) break;
    /* correct band? */
    else if (line[0]!=band) {iline++; continue;}
    /* right band */
    gotBand = TRUE;

    /* Parse antenna, start and end dates for line */
    n = sscanf(&line[1], "%lf %d %d %d", &tfreq, &antno, &sdate, &edate);
    /* New frequency? */
    if ((iarray<0) || (farray[iarray]<=0.0)) {
      iarray++;
      narray++;
      farray[iarray] = tfreq;
      /* Use default? */
      if ((narray>1) && (garray[iarray][0]<=0.0)) {
	garray[iarray][0] = avgGain[0];
	garray[iarray][1] = avgGain[1];
	garray[iarray][2] = avgGain[2];
	garray[iarray][3] = avgGain[3];
	/* Still undefined? use flat */
	if (garray[iarray][0]<=0.0) {
	  garray[iarray][0] = 1.0;
	  garray[iarray][1] = 0.0;
	  garray[iarray][2] = 0.0;
	  garray[iarray][3] = 0.0;
	}
      }
      avgGain[0] = avgGain[1] = avgGain[2] = avgGain[3] = 0.0;
    }
      
    /* Save average parameters */
    if (antno==0) {
      /* parse reference ant values */
      n = sscanf(&line[30], "%f %f %f %f", 
		 &avgGain[0], &avgGain[1], &avgGain[2], &avgGain[3]);
    }  /* end average antenna */

    /* Start past JD? */
    if (sdate>date) break;
    /* Right antenna and In range? */
    if ((Ant==antno) && (date>sdate) && (date<edate)) {
      /* Got it */
      found = TRUE;
      /* parse coefficients */
      n = sscanf(&line[30], "%f %f %f %f", 
		 &garray[iarray][0], &garray[iarray][1], 
		 &garray[iarray][2], &garray[iarray][3]);
    }
    iline++;    /* next entry */
  } /* end loop over table */

  /* Use default for last gain? */
  if ((narray>=1) && (garray[iarray][0]<=0.0)) {
    garray[iarray][0] = avgGain[0];
    garray[iarray][1] = avgGain[1];
    garray[iarray][2] = avgGain[2];
    garray[iarray][3] = avgGain[3];
    /* Still undefined? use flat */
    if (garray[iarray][0]<=0.0) {
      garray[iarray][0] = 1.0;
      garray[iarray][1] = 0.0;
      garray[iarray][2] = 0.0;
      garray[iarray][3] = 0.0;
    }
  }

  /* Get interpolated curves */
  if (narray>=1) {
    /* have gain curves */
    GetGainCurves (4, narray, farray, garray, nfreq, gainFreq, gain);
  } else {
    /* Nothing - use flat curves */
    for (i=0; i<nfreq; i++) {
      gain[i][0] = 1.0;
      gain[i][1] = gain[i][2] = gain[i][3] = 0.0;
    }
  }

  /* Cleanup */
  if (garray) {
    for (i=0; i<8; i++) if (garray[i]) g_free(garray[i]);
    g_free(garray);
  }
} /* end ObitVLAGainParseGain */
